/*
function T=cpd_transform(Z, Transform)

switch lower(Transform.method)
    case {'rigid','affine'}
        T=Transform.s*(Z*Transform.R')+repmat(Transform.t',[size(Z,1) 1]);
    case {'nonrigid'}
        
        % because of normalization during the registration
        % these steps are necessary for non-rigid tranaformation
        Transform.beta=Transform.beta*Transform.normal.yscale;
        Transform.W=Transform.normal.xscale*Transform.W;
        Transform.shift=Transform.normal.xd-Transform.normal.xscale/Transform.normal.yscale*Transform.normal.yd;
        Transform.s=Transform.normal.xscale/Transform.normal.yscale;

        G=cpd_G(Z, Transform.Yorig,Transform.beta);
        T=Z*Transform.s+G*Transform.W+repmat(Transform.shift,size(Z,1),1);
        % T=Transform.s*Z+repmat(Transform.t',[size(Z,1) 1])+cpd_G(Z, Transform.Yorig,Transform.beta)*Transform.W;
    otherwise
        error('CPD: This transformation is not supported.')
end         
*/


/** @file	cpd_transform.cpp
 *  @author	Herve Lombaert
 *  @date	March 9th, 2011
 *  @brief	C version of cpd_transform.m
 */

#include <mex.h>
#include <matrix.h>
#include <cstring>
#include <cmath>

#define TRY try {
#define CATCH } catch(const char* msg) { mexErrMsgTxt(msg); } catch(...) { mexErrMsgTxt("Unknown Error\n"); }


double* cpd_transform_rigid(const double* Z, unsigned int nbPts, unsigned int nbDims,
                            double s, const double* R, const double* t)
{
    double* T = new double [nbDims*nbPts]; // T[component*nbPts + ptNb]
    for(unsigned int i=0; i!=nbDims*nbPts; ++i) T[i] = 0;
    
    /* Matlab Code:
     *
        T=Transform.s*(Z*Transform.R')+repmat(Transform.t',[size(Z,1) 1]);
     */
    for(unsigned int pt=0; pt!=nbPts; ++pt)
    {
        for(unsigned int d=0; d!=nbDims; ++d)
        {
            /* rotation */
            for(unsigned int i=0; i!=nbDims; ++i)
                T[d*nbPts + pt] += s*Z[i*nbPts+pt]*R[d*nbDims+i];
            
            /* translation */
            T[d*nbPts + pt] += t[d];
        }
    }
    
    return T;
}

double* cpd_transform_nonrigid(const double* Z, unsigned int nbPts, unsigned int nbDims,
                               unsigned int nbCtlPts, double beta, const double* W, const double* Yorig, double s,
                               const double* xd, const double* yd, double xscale, double yscale)
{
    double* T = new double [nbDims*nbPts]; // T[component*nbPts + ptNb]
    for(unsigned int i=0; i!=nbDims*nbPts; ++i) T[i] = 0;
    
    /* Matlab Code:
     *
        Transform.beta=Transform.beta*Transform.normal.yscale;
        Transform.W=Transform.normal.xscale*Transform.W;
        Transform.shift=Transform.normal.xd-Transform.normal.xscale/Transform.normal.yscale*Transform.normal.yd;
        Transform.s=Transform.normal.xscale/Transform.normal.yscale;
     */
    beta = beta * yscale;
    double* sW = new double [nbCtlPts*nbDims];
    for(unsigned int i=0; i!=nbCtlPts*nbDims; ++i)
        sW[i] = xscale*W[i];
    double* shift = new double [nbDims];
    for(unsigned int i=0; i!=nbDims; ++i)
        shift[i] = xd[i] - xscale/yscale * yd[i];
    s = xscale / yscale;
            
    /* Matlab Code:
     *
        G=cpd_G(Z, Transform.Yorig,Transform.beta);
        T=Z*Transform.s+G*Transform.W+repmat(Transform.shift,size(Z,1),1);
     *
     * where:
     *
        function G=cpd_G(x,y,beta)

        if nargin<3, error('cpd_G.m error! Not enough input parameters.'); end;

        k=-2*beta^2;
        [n, d]=size(x); [m, d]=size(y);

        % G=repmat(x,[1 1 m])-permute(repmat(y,[1 1 n]),[3 2 1]);
        % G=squeeze(sum(G.^2,2));
        % G=G/k;
        % G=exp(G);

        % Avoiding using repmat to save memory - however the results are just to
        % big to store them in G(n,m)
        G = sparse(n,m,0);
        for i=1:m
            diff = bsxfun(@minus,x,y(i,:));
            G(:,i) = sum(diff.^2,2);
        end
        G=G/k;
        G=exp(G);
     */
    double* g    = new double [nbCtlPts];
    double factor = -1 / (2*beta*beta);
    for(unsigned int pt=0; pt!=nbPts; ++pt)
    {
        // T=Z*Transform.s + repmat(Transform.shift,size(Z,1),1);
        for(unsigned int d=0; d!=nbDims; ++d)
            T[d*nbPts + pt] = s*Z[d*nbPts + pt] + shift[d];
        
        // g(i) = exp(-|pt-yorig(i)|^2/2beta^2): one row of nbCtlPts gaussian weights
        for(unsigned int i=0; i!=nbCtlPts; ++i)
        {
            double norm2 = 0;
            for(unsigned int d=0; d!=nbDims; ++d)
            {
                double diff = Z[d*nbPts + pt] - Yorig[d*nbCtlPts + i];
                norm2 += diff*diff;
            }
                
            g[i] = exp(norm2*factor);
        }
        
        // T = T + g(i)*W(i)
        for(unsigned int d=0; d!=nbDims; ++d)
        {
            for(unsigned int i=0; i!=nbCtlPts; ++i)
            {
                T[d*nbPts + pt] += g[i]*sW[d*nbCtlPts + i];
            }
        }
    }
    delete [] g; g = 0;
    
    return T;
}

void mexFunction(int nlhs, mxArray* plhs[],
                 int nrhs, const mxArray *prhs[])
{
	TRY;

	if(nrhs<2)  { mexPrintf("Need to provide points and a transformation\n"); return; }
    if(!mxIsStruct(prhs[1])) { mexPrintf("Invalid transformation\n"); return; }

    int            n=0,m=0;         // array sizes
    double*        data =0;         // array data
    const mxArray* array =0;        // matlab array

	/* Get points to warp */
    unsigned int nbPts        = mxGetM(prhs[0]);
    unsigned int nbDims       = mxGetN(prhs[0]);
    double* Z                 = mxGetPr(prhs[0]);
	//mexPrintf("Size:       (%d,%d)\n", nbPts, nbDim);
    
    /* Get the transformation type */
    array = mxGetField(prhs[1],0,"method"); if(!array)    { mexPrintf("No method\n"); return; }
    n = mxGetN(array)+1;
    char* str = new char [n];
	if(mxGetString(array, str, n)!=0)                     { mexPrintf("Invalid transformation\n"); return; }
    //mexPrintf("Method:     %s\n", str);
    
    enum METHOD { RIGID, NONRIGID, OTHER } method = OTHER;
    
    if(!strncmp(str,"rigid",5) || !strncmp(str,"affine",6))
        method = RIGID;
    if(!strncmp(str,"nonrigid",8))
        method = NONRIGID;

    if(str) { delete [] str; str = 0; }

    /* Transformation variables */
    unsigned int nbCtlPts     = 0;
    double beta               =0;
    double* Y                 =0;
    double* W                 =0;
    double* Yorig             =0;
    double s                  =0;
    double* R                 =0;
    double* t                 =0;
    double* xd                =0;
    double* yd                =0;
    double xscale             =0;
    double yscale             =0;
    
    /* Get transformation */
    if(method == NONRIGID)
    {
        array = mxGetField(prhs[1],0,"Y");      if(!array)      { mexPrintf("No Y\n"); return; }
        nbCtlPts = mxGetM(array);
        n = mxGetN(array);                      if(n!=nbDims)   { mexPrintf("Y wrong dimension\n"); return; }
        Y = mxGetPr(array);                     if(!Y)          { mexPrintf("Invalid Y\n"); return; } // Y[component*sizeW[0] + ptNb]
    }
    
    if(method == NONRIGID)
    {
        array = mxGetField(prhs[1],0,"beta");   if(!array)      { mexPrintf("No beta\n"); return; }
        m = mxGetM(array);                      if(m!=1)        { mexPrintf("Invalid beta\n"); return; }
        n = mxGetN(array);                      if(n!=1)        { mexPrintf("Invalid beta\n"); return; }
        data = mxGetPr(array);                  if(!data)       { mexPrintf("Invalid beta\n"); return; }
        beta = data[0];
        //mexPrintf("Beta:       %f\n", beta);

        array = mxGetField(prhs[1],0,"W");      if(!array)      { mexPrintf("No W\n"); return; }
        m = mxGetM(array);                      if(m!=nbCtlPts) { mexPrintf("W wrong dimension\n"); return; }
        n = mxGetN(array);                      if(n!=nbDims)   { mexPrintf("W wrong dimension\n"); return; }
        W = mxGetPr(array);                     if(!W)          { mexPrintf("Invalid W\n"); return; } // W[component*sizeW[0] + ptNb]

        array = mxGetField(prhs[1],0,"Yorig");  if(!array)      { mexPrintf("No Yorig\n"); return; }
        m = mxGetM(array);                      if(m!=nbCtlPts) { mexPrintf("Yorig wrong dimension\n"); return; }
        n = mxGetN(array);                      if(n!=nbDims)   { mexPrintf("Yorig wrong dimension\n"); return; }
        Yorig = mxGetPr(array);                 if(!Yorig)      { mexPrintf("Invalid Yorig\n"); return; } // Yorig[component*sizeW[0] + ptNb]
    }
    
    if(method == NONRIGID || method == RIGID)
    {
        array = mxGetField(prhs[1],0,"s");      if(!array)      { mexPrintf("No s\n"); return; }
        m = mxGetM(array);                      if(m!=1)        { mexPrintf("Invalid s\n"); return; }
        n = mxGetN(array);                      if(n!=1)        { mexPrintf("Invalid s\n"); return; }
        data = mxGetPr(array);                  if(!data)       { mexPrintf("Invalid s\n"); return; }
        s = data[0];
        //mexPrintf("s:          %f\n", s);
    }
    
    if(method == RIGID)
    {
        array = mxGetField(prhs[1],0,"R");      if(!array)      { mexPrintf("No R\n"); return; }
        m = mxGetM(array);                      if(m!=nbDims)   { mexPrintf("R wrong dimension\n"); return; }
        n = mxGetN(array);                      if(n!=nbDims)   { mexPrintf("R wrong dimension\n"); return; }
        R = mxGetPr(array);                     if(!R)          { mexPrintf("Invalid R\n"); return; }
    }
    
    if(method == NONRIGID || method == RIGID)
    {
        array = mxGetField(prhs[1],0,"t");      if(!array)      { mexPrintf("No t\n"); return; }
        m = mxGetM(array);                      if(m!=nbDims)   { mexPrintf("t wrong dimension\n"); return; }
        n = mxGetN(array);                      if(n!=1)        { mexPrintf("t wrong dimension\n"); return; }
        t = mxGetPr(array);                     if(!t)          { mexPrintf("Invalid t\n"); return; }
    }
    
    if(method == NONRIGID)
    {
        const mxArray* normal = mxGetField(prhs[1],0,"normal"); if(!normal) { mexPrintf("No normal\n"); return; }
        if(!mxIsStruct(normal))                                 { mexPrintf("Invalid normal\n"); return; }

        array = mxGetField(normal,0,"xd");      if(!array)      { mexPrintf("No xd\n"); return; }
        m = mxGetM(array);                      if(m!=1)        { mexPrintf("Invalid xd\n"); return; }
        n = mxGetN(array);                      if(n!=nbDims)   { mexPrintf("xd wrong dimension\n"); return; }
        xd = mxGetPr(array);                    if(!xd)         { mexPrintf("Invalid xd\n"); return; } // xd[component]

        array = mxGetField(normal,0,"yd");      if(!array)      { mexPrintf("No yd\n"); return; }
        m = mxGetM(array);                      if(m!=1)        { mexPrintf("Invalid yd\n"); return; }
        n = mxGetN(array);                      if(n!=nbDims)   { mexPrintf("yd wrong dimension\n"); return; }
        yd = mxGetPr(array);                    if(!yd)         { mexPrintf("Invalid yd\n"); return; } // yd[component]

        array = mxGetField(normal,0,"xscale");  if(!array)      { mexPrintf("No xscale\n"); return; }
        m = mxGetM(array);                      if(m!=1)        { mexPrintf("Invalid xscale\n"); return; }
        n = mxGetN(array);                      if(n!=1)        { mexPrintf("Invalid xscale\n"); return; }
        data = mxGetPr(array);                  if(!data)       { mexPrintf("Invalid xscale\n"); return; }
        xscale = data[0];

        array = mxGetField(normal,0,"yscale");  if(!array)      { mexPrintf("No yscale\n"); return; }
        m = mxGetM(array);                      if(m!=1)        { mexPrintf("Invalid yscale\n"); return; }
        n = mxGetN(array);                      if(n!=1)        { mexPrintf("Invalid yscale\n"); return; }
        data = mxGetPr(array);                  if(!data)       { mexPrintf("Invalid yscale\n"); return; }
        yscale = data[0];
    }

    /* Warp points */
    double* T = 0; // the transformed points
    
    switch(method)
    {
        case RIGID:
            T = cpd_transform_rigid(Z,nbPts,nbDims, s,R,t);
            break;
        case NONRIGID:
            T = cpd_transform_nonrigid(Z,nbPts,nbDims, nbCtlPts,beta,W,Yorig,s, xd,yd,xscale,yscale);
            break;
        default:
            mexPrintf("CPD: This transformation is not supported.");
            break;
    }
    
    if(T)
    {
        plhs[0] = mxCreateDoubleMatrix(nbPts,nbDims,mxREAL); // the transformed points
        double* data = mxGetPr(plhs[0]);
        
        for(unsigned int i=0; i!=nbPts*nbDims; ++i)
            data[i] = T[i];
    }

    CATCH;
}
