% ADJACENCY Build an adjacency matrix
%
%    W = ADJACENCY(X,F,metric,k) builds an adjacency matrix of a mesh.
%    The mesh is defined with its vertices X and its faces F.
%    The edge weights are defined by default using L2 distances
%
%    if F is not provided or empty (e.g., unknown face structure),
%    edges are built using the k nearest neighbors of each vertex.
%
%    metric can be either:
%       'L2'     - weights edges using L2 distances (default)
%       'gauss'  - weights edges using a gaussian kernel where
%                  sigma=max(dist)/3 (using the max of all edge lengths)
%       'one'    - weights edges using only ones
%                  creates an unweighted adjacency matrix
%
%    Example:
%       [X,F] = read_mesh('cortex-0');
%
%       % finds the adjacency matrix of the read mesh
%       W = adjacency(X,F);
%
%       % finds the adjacency matrix using K nearest neighbors
%       % and a gaussian kernel for edge weights
%       W = adjacency(X,[],'gauss');
%
%    See also eigenmap

% file:    adjacency.m
% author:  Herve Lombaert
% contact: herve.lombaert@gmail.com
% date:    April 16th, 2011

function W = adjacency(X,F,metric,k)

if nargin<3; metric = 'L2'; end;
if nargin<4; k      = 12;   end;

%% Find neighboring points
if nargin>1 && ~isempty(F)
    
    % In the triangle face defined with v1,v2,v3:
    % links are v1->v2,v1->v3, v2->v1,v2->v3, v3->v1,v3->v2

    idx_org = [F(:,1); F(:,1); F(:,2); F(:,2); F(:,3); F(:,3)]; % edge origin
    idx_end = [F(:,2); F(:,3); F(:,1); F(:,3); F(:,1); F(:,2)]; % edge end
    N = [idx_org idx_end];              % N(:,1) are edge origins, N(:,2) edge ends
    N = unique(sort(N,2),'rows');       % remove duplicates
    
else
    
    % Link all points to their K nearest neighbors

    N = knn(X,k);                       % N(:,1) are edge origins, N(:,2) edge ends
    N = [repmat((1:size(X,1))',k,1) N(:)];
    N = [N(:,1) N(:,2); N(:,2) N(:,1)]; % add reverse edges
    N = unique(sort(N,2),'rows');       % remove duplicates

end
N = double(N);

%% Compute edge weights
switch metric

    case 'L2'
        % compute length of all links (euclidean distance)
        dist    = [sum((X(N(:,1),:)-X(N(:,2),:)).^2,2)].^0.5;
        eps     = 1e-6; % penalize small distances (avoid division by zero)
        weights = (dist+eps).^-1;
        weights(weights>1000) = 1000; % trunc max weight

    case 'gauss'
        % compute weights for all links (gaussian kernel)
        dist2   = [sum((X(N(:,1),:)-X(N(:,2),:)).^2,2)];
        sigma   = max(dist2)/3; % uses the max of all distances
        weights = exp(-dist2.^2 ./ (2*sigma^2));

    case 'one'
        % edge lenghts are all ones
        weights = ones(size(N,1),1); % unweighted adjacency matrix

    otherwise
        error(['Unknwon metric: ' metric]);
end
    
%% Fill Adjacency matrix
W = sparse(N(:,1), N(:,2), weights, size(X,1),size(X,1));
W = (W+W')/2; % correct any non symmetry (this will avoid any error in eigendecomposition)
