% EIGENMAP Compute spectral components of a mesh
%
%    [E,V] = EIGENMAP(X,F,option) computes the eigenvectors of the graph
%    Laplacian that describes the given mesh. The mesh is defined with its
%    vertices, X, and its triangular face structure, F. If F is not
%    provided, the neighborhood connections are defined using
%    k-nearest-neighbors. If option is a number, it is the number of
%    non-null eigenvectors to compute. If it is a structure, the parameters
%    are described below. The graph Laplacian can use extra features on
%    the weighting of graph edges, which defines the inter-nodal
%    relationship within the mesh, and on graph nodes, which defines the
%    importance of each node. The outputs are the sorted eigenvectors
%    (one eigenvector per column) and the eigenvalues. The null components
%    are ignored, and the first column is always the Fiedler vector.
%
%    Optional parameters are:
%       option.ne     - number of non-null eigenvectors      (default: 9)
%       option.EW     - feature to use in graph edge weights (default: [])
%       option.NW     - feature to use in graph node weights (default: [])
%       option.method - edge weight method can either be 'add' (default),
%                       for additive where the adjacency matrix is:
%                          adjacency([X f*EW],F), with
%                          f=gamma*range(X)/range(EW)
%                       or 'mult' for multiplicative where the adjacency
%                       matrix is:
%                          adjacency(X,F) .* adjacency(EW,F)
%       option.metric - edge weight can either be 'one', 'L2' (default),
%                       or 'gauss'. See adjacency
%       option.knn    - k nearest neighbors if no face struct(default: 12)
%       option.gamma  - factor used when the method is 'add' (default: 0.2)
%       option.alpha  - factor used on node weights          (default: 1.0)
%
%    Example 1:
%       [X,F] = read_mesh('cortex-0');
%
%       % compute five spectral components
%       [E,V] = eigenmap(X,F,5);
%
%    Example 2:
%       S = read_sulcaldepth('cortex-1');       % extra feature
%
%       % compute five spectral components
%       option.ne = 5;
%       option.EW = S;
%       [E,V] = eigenmap(X,F,option);
%
%    See also adjacency match reordereigenmaps
%
%    Reference:
%    [1] Spectral correspondence for Brain Matching,
%        Lombaert, Grady, Polimeni, Cheriet, IPMI'11

% file:    eigenmap.m
% author:  Herve Lombaert
% contact: herve.lombaert@gmail.com
% date:    April 16th, 2011

function [E,V] = eigenmap(X,F,option)

%% Default parameters
ne       = 9;      % number of eigenvectors to use
EW       = [];     % extra features on edge weights
NW       = [];     % extra features on node weights
gamma    = 0.2;    % scaling factor for edge weighting
alpha    = 1;      % scaling factor for node weighting
method   = 'add';  % type of edge weight
metric   = 'L2';   % metric of edge weight
knn      = 12;     % number of nearest neighbors
verbose  = 1;      % display timing

if nargin<2; F = []; end;

if nargin==3
    if isstruct(option)
        if isfield(option,'ne');        ne      = option.ne;       end
        if isfield(option,'EW');        EW      = option.EW;       end
        if isfield(option,'NW');        NW      = option.NW;       end
        if isfield(option,'gamma');     gamma   = option.gamma;    end
        if isfield(option,'alpha');     alpha   = option.alpha;    end
        if isfield(option,'method');    method  = option.method;   end
        if isfield(option,'metric');    metric  = option.metric;   end
        if isfield(option,'knn');       knn     = option.knn;      end
        if isfield(option,'verbose');   verbose = option.verbose;  end
    else
        ne = option; % if option is a number, it is nb of eigenvectors
    end
end

% failsage
if isequal(size(gamma),[1 1]); gamma = gamma * ones(1,size(EW,2)); end;
if isequal(size(alpha),[1 1]); alpha = alpha * ones(1,size(NW,2)); end;

switch method

    case 'add'
        %% Adding extra coordinate to mesh
        maxrangeX = max(range(X));
        for i=1:size(EW,2)
            factor = gamma(i) * maxrangeX / range(EW(:,i));
            %factor = gamma(i) * range(EW(:,i)) / max(range(X));
            X = [X factor*EW(:,i)];
        end

        %% Adjacency matrix
        W = adjacency(X,F,metric,knn);

    case 'mult'
        %% Adjacency matrix
        W = adjacency(X,F,metric,knn);
        
        for i=1:size(EW,2)
            W = W .* adjacency(EW(:,i),F,metric,knn);
        end
        
    otherwise
        error(['Unknwon method: ' method]);
end

%% Laplacian
D    = spdiags(sum(W)',     0,speye(size(W,1))); %D = diag(sum(W));
Dinv = spdiags(sum(W)'.^-1, 0,speye(size(W,1))); %Dinv = D^-1;
L    = D - W;

%% Node weighting
if size(NW,2)>0
    G = 0;
    for i=1:size(NW,2)
        NW(:,i) = exp(NW(:,i));                              % positive weighting
        factor = alpha(i) * range(diag(D)) / range(NW(:,i)); % normalization
        G      = G + factor * NW(:,i);                       % adding node weight
    end
    G = G / size(NW,2);
    G = spdiags(G.^1,0,speye(size(W,1)));                    % diagonalize it
    G = Dinv * G;                                            % total node weight
else
    G = Dinv; % node weight is a neighborhood weighting
end

%% Normalized Laplacian
L = G * L;

%% Compute eigen maps
lastwarn('');
o.disp   = 0;
o.isreal = 1;
o.issym  = 0;
o.maxit  = 50;
[E,V] = eigs(L,ne+1,1e-10,o); % arpack eigen decomposition
V = diag(V);
if ~isempty(lastwarn); error(lastwarn); end

E = real(E); % safety net
V = real(V);

%% Sort eigen vectors
[V,idx] = sort(V);
E = E(:,idx);

%% Remove null vector
E = E(:,2:end);
V = V(2:end);

%return
%% Find Fiedler vector (first non zero eigen component)
fiedler = 0;
for i=1:size(V,1); if V(i)>1e-10; fiedler = i; break; end; end;
if fiedler<1; error('No Fiedler vector'); end;
if fiedler>1;
    if verbose; disp(['Fiedler vector is ' num2str(fiedler)]); end;
    E = E(:,fiedler:end);
    V = V(:,fiedler:end);
    if size(E,2)<ne;
        lastwarn('');
        [E,V] = eigs(L,ne+fiedler,1e-10,o); % arpack eigen decomposition
        V = diag(V);
        if ~isempty(lastwarn); error(lastwarn); end

        E = real(E); % safety net
        V = real(V);

        [V,idx] = sort(V);
        E = E(:,idx);

        E = E(:,(1+fiedler):end);
        V = V((1+fiedler):end);
    end
end;

end
