% MATCH Find the correspondence map
%
%    [CORR,CORR_INV,E1,E2] = MATCH(X1,X2,F1,F2,option)
%    finds the one-to-one correspondence map projecting nodes from
%    mesh 1 to mesh 2. The meshes are defined with their vertices, X1, X2,
%    and its faces F. Extra features on both meshes, EX1, EX2 (one feature
%    per column) are optional. These extra features are used to match
%    nodes, and they are spatially regularized using spectral components.
%    Spectral components are generated from the mesh structure. Spectral
%    components can be customized with optional edge weighting, EW1, EW2
%    (one feature per column), controlling the neighborhood relationship.
%    Spectral components can also be customized with optional node
%    weighting, NW1, NW2 (one feature per column), controlling the
%    importance of each node. Optional outputs, CORR_INV, is the inverse
%    transformation from mesh 2 to mesh 1, and E1, E2, are the regularized
%    spectral components (one component per column).
%
%    Optional parameters can be given:
%       option.ne       - number of spectral components      (default: 9)
%       option.beta     - factors used in the extra features (default: 1.0)
%       option.gamma    - factors used in the edge weights   (default: 0.2)
%       option.alpha    - factors used in the node weights   (default: 1.0)
%       option.method   - edge weight method can either be 'add' (default),
%                         for additive where the adjacency matrix is:
%                            adjacency([X f*EW],F), with
%                            f=gamma*range(X)/range(EW)
%                         or 'mult' for multiplicative where the adjacency
%                         matrix is:
%                            adjacency(X,F).*adjacency(EW,F)
%       option.metric   - edge weight can either be 'one', 'L2' (default),
%                         or 'gauss'. See adjacency
%       option.ex_smooth- smoothing level of extra features during matching
%                                                            (default: 30)
%       option.cpd_nctl - number of control points during alignment
%                                                            (default 1500)
%       option.eta      - factors used on extra features when finding
%                         closest embedding points           (default 1.0)
%       option.sm_level - smoothing level of the final correspondence map
%                         using 0 disable this smoothness    (default: 40)
%       option.sm_mesh  - mesh smooth level. The final correspondence map
%                         can be smoothed on a smoothed mesh (default: 300)
%       option.verbose  - display timing                     (default: 1)
%
%    Internally, all features are normalized between 0 and 1. The optional
%    factors, beta, gamma, alpha, can change the scale of this
%    normalization. The final correspondence map is smoothed using a
%    spatial filter on a smoothed mesh. The smoothness level of the filter
%    is controled with sm_level, and the smoothness of the mesh is
%    controled with sm_mesh.
%
%    Example 1:
%       [X1,F1] = read_mesh('cortex-1');
%       [X2,F2] = read_mesh('cortex-2');
%
%       % finds the correspondence map between mesh 1 and mesh 2
%       corr = match(X1,X2,F1,F2);
%
%       displaycorrespondence(X1,X2,F1,F2,corr); % display map
%
%    Example 2:
%       S1 = read_sulcaldepth('cortex-1');       % first extra feature
%       T1 = read_corticalthickness('cortex-1'); % second extra feature
%       S2 = read_sulcaldepth('cortex-2');
%       T2 = read_corticalthickness('cortex-2');
%
%       % finds the correspondence map between mesh 1 and mesh 2
%       % using extra features and using them on edge weights
%       option.EX1 = [S1 T1];
%       option.EX2 = [S2 T2];
%       option.EW1 = [S1 T1];
%       option.EW2 = [S2 T2];
%       option.NW1 = [S1 T1];
%       option.NW2 = [S2 T2];
%       corr = match(X1,X2,F1,F2,option);
%
%       displaycorrespondence(X1,X2,F1,F2,corr); % display map
%
%    See also eigenmap adjacency aligneigenmaps reordereigenmaps
%
%    Reference:
%    [1] Spectral correspondence for Brain Matching,
%        Lombaert, Grady, Polimeni, Cheriet, IPMI'11

% file:    adjacency.m
% author:  Herve Lombaert
% contact: herve.lombaert@gmail.com
% date:    April 16th, 2011


function [corr,corr_inv,E1,E2] = match(X1,X2,F1,F2,option)

%% Default Parameters
if nargin<4;  F1 =[]; F2 =[];  end; % no face structure
corr_inv = [];
EX1      = [];   % extra features on CPD
EX2      = [];   % extra features on CPD
EW1      = [];   % extra features on edge weighting
EW2      = [];   % extra features on edge weighting
NW1      = [];   % extra features on node weighting
NW2      = [];   % extra features on node weighting
ne       = 9;    % number of eigenvectors to use
beta     = 1;    % scaling factor for extra features on CPD
gamma    = 1;    % scaling factor for edge weighting
alpha    = 1;    % scaling factor for node weighting
method= 'add';   % type of edge weight
metric= 'L2';    % metric of edge weight
cpd_nctl = 1500; % nb of control points to use in CPD
eta      = 1;    % weights of features in embeddings after CPD
ex_smooth= 30;   % smooth extra features during CPD
sm_level = 40;   % filter iteration nb. during map smoothing
sm_mesh  = 300;  % smooth level of mesh during map smoothing
verbose  = 1;    % display timing

if nargin==5 &&  isstruct(option)
    if isfield(option,'EX1');       EX1       = option.EX1;        end;
    if isfield(option,'EX2');       EX2       = option.EX2;        end;
    if isfield(option,'EW1');       EW1       = option.EW1;        end;
    if isfield(option,'EW2');       EW2       = option.EW2;        end;
    if isfield(option,'NW1');       NW1       = option.NW1;        end;
    if isfield(option,'NW2');       NW2       = option.NW2;        end;
    if isfield(option,'ne');        ne        = option.ne;         end;
    if isfield(option,'beta');      beta      = option.beta;       end;
    if isfield(option,'gamma');     gamma     = option.gamma;      end;
    if isfield(option,'alpha');     alpha     = option.alpha;      end;
    if isfield(option,'method');    method    = option.method;     end;
    if isfield(option,'metric');    metric    = option.metric;     end;
    if isfield(option,'cpd_nctl');  cpd_nctl  = option.cpd_nctl;   end;
    if isfield(option,'eta');       eta       = option.eta;        end;
    if isfield(option,'ex_smooth'); ex_smooth = option.ex_smooth;  end;
    if isfield(option,'sm_level');  sm_level  = option.sm_level;   end;
    if isfield(option,'sm_mesh');   sm_mesh   = option.sm_mesh;    end;
    if isfield(option,'verbose');   verbose   = option.verbose;    end;
end

% failsafe
if size(X1,2) ~=size(X2,2);  error('X1 and X2 must have same dimensions');   end;
if size(EX1,2)~=size(EX2,2); error('EX1 and EX2 must have same dimensions'); end;
if size(EW1,2)~=size(EW2,2); error('EW1 and EW2 must have same dimensions'); end;
if size(NW1,2)~=size(NW2,2); error('NW1 and NW2 must have same dimensions'); end;

if isequal(size(beta), [1 1]); beta  = beta  * ones(1,size(EX1,2)); end;
if isequal(size(gamma),[1 1]); gamma = gamma * ones(1,size(EX1,2)); end;
if isequal(size(alpha),[1 1]); alpha = alpha * ones(1,size(EX1,2)); end;

if verbose; total_t = cputime; end;

if ne>1
    
    %% Compute eigen maps
    if verbose; disp(['Computing Eigenmaps 1']); tic;        end;
    
    o.EW      = EW1;
    o.NW      = NW1;
    o.ne      = ne;
    o.method  = method;
    o.metric  = metric;
    o.gamma   = gamma;
    o.alpha   = alpha;
    o.verbose = verbose;
    [E1,V1] = eigenmap(X1,F1,o);

    if verbose; disp(['Computing Eigenmaps 2']);             end;
    o.EW      = EW2;
    o.NW      = NW2;
    o.ne      = ne;
    o.method  = method;
    o.metric  = metric;
    o.gamma   = gamma;
    o.alpha   = alpha;
    o.verbose = verbose;
    [E2,V2] = eigenmap(X2,F2,o);
    
    if verbose; t = toc; disp(['Done: ' num2str(t) ' sec']); end;

    %% Reorder and weight eigenmaps
    if verbose; disp(['Reordering Eigenmaps']); tic; end;
    [E1,E2,Q] = reordereigenmaps(X1,X2,E1,E2,V1,V2); % Q is the permutation cost

    % Weight eigenvectors
    weight  = Q.*max(V1,V2);
    sigma   = mean(weight);
    weight  = exp(-(weight.^2)/(2*sigma^2)); % high frequencies have less importance
    E1      = bsxfun(@times,E1,weight');
    E2      = bsxfun(@times,E2,weight');
    rangeE1 = range(E1(:,1));
    rangeE2 = range(E2(:,1));

    if verbose; t = toc; disp(['Done: ' num2str(t) ' sec']); end;
    
else
    rangeE1 = 1;
    rangeE2 = 1;
end

%% Align using CPD
Y1 = []; % embedding
Y2 = []; % embedding

% Smooth and embed extra features
W1 = adjacency(X1,F1,'one');
W2 = adjacency(X2,F2,'one');

for i=1:size(EX1,2)
    EX1(:,i) = filtergraph(W1,EX1(:,i),ex_smooth);      % filter extra feature
    EX2(:,i) = filtergraph(W2,EX2(:,i),ex_smooth);
    nEX1 = EX1(:,i) / range(EX1(:,i)) * rangeE1 * beta; % normalize with respect to the eigenvectors
    nEX2 = EX2(:,i) / range(EX2(:,i)) * rangeE2 * beta;
    Y1 = [nEX1 Y1]; % Add extra feature to the embedding
    Y2 = [nEX2 Y2];
end

% Concatenate eigen vectors
Y1 = [Y1 E1];
Y2 = [Y2 E2];

% Align using CPD
if verbose; disp(['Aligning Eigenmaps']); tic;           end;
[Y1,Y2] = aligneigenmaps(Y1,Y2,cpd_nctl);

if verbose; t = toc; disp(['Done: ' num2str(t) ' sec']); end;

% Retain aligned eigenvectors
E1 = Y1(:,(end-ne+1):end);
E2 = Y2(:,(end-ne+1):end);

%% Find closest points on embeddings
Y1(:,1:(end-ne)) = eta * Y1(:,1:(end-ne)); % weight the extra coordinates
Y2(:,1:(end-ne)) = eta * Y2(:,1:(end-ne));

if verbose; disp(['Finding Pairs']); tic;                end;
tree = kdtree(Y2);
corr = kdtree_closestpoint(tree, Y1); % all corresponding point on embeddings
% corr = zeros(size(X2,1),1);
% for iter=1:size(X2,1)
%     d2 = sum(bsxfun(@minus,Y1,Y2(iter,:)).^2,2);
%     [m,idx] = min(d2);
%     corr(iter) = idx;
% end

if nargout>1
    tree = kdtree(Y1);
    corr_inv = kdtree_closestpoint(tree,Y2); % all corresponding point on embeddings
end

if verbose; t = toc; disp(['Done: ' num2str(t) ' sec']); end;

%% Smooth correspondence map
if sm_level>0
    
    if verbose; tic;     disp(['Smooth Map']);               end;
    if ~isempty(F2)
        sX2 = smoothmesh(X2,F2,sm_mesh);                          % smooth mesh 2
    else
        sX2 = filtergraph(W2,X2,sm_mesh);                         % smooth mesh 2
    end
    X1_proj = filtergraph(W1,sX2(corr,:),sm_level);               % smoothed coordinates of projection of X1 on mesh 2

    % Find the closest cartesian point to the smoothed corresponding point
    tree = kdtree(sX2);
    corr_smooth = kdtree_closestpoint(tree, X1_proj);             % all corresponding point on mesh 2

    if verbose; t = toc; disp(['Done: ' num2str(t) ' sec']); end;

    corr = corr_smooth;

    if nargout>1
        if ~isempty(F1)
            sX1 = smoothmesh(X1,F1,sm_mesh);                          % smooth mesh 1
        else
            sX1 = filtergraph(W1,X1,sm_mesh);                         % smooth mesh 1
        end
        X2_proj = filtergraph(W2,sX1(corr_inv,:),sm_level);               % smoothed coordinates of projection of X2 on mesh 1

        % Find the closest cartesian point to the smoothed corresponding point
        tree = kdtree(sX1);
        corr_inv_smooth = kdtree_closestpoint(tree, X2_proj);             % all corresponding point on mesh 1
        
        corr_inv = corr_inv_smooth;
    end
end

if verbose; disp(['Total Time: ' num2str(cputime-total_t) ' sec']); end;

