% REORDEREIGENMAPS Reorder an eigenmap
%
%    [E1,E2,Q] = REORDEREIGENMAPS(X1,X2,E1,E2,V1,V2) reorders and flip
%    eigenvector signs of the second eigenmap so its coordinates are
%    equivalent with the coordinates of the first eigenmap. The output, Q,
%    is the permutation cost of the ith eigenvector pair.
%
%    Example:
%       [E1,V1] = eigenmap(X1,F1);
%       [E2,V2] = eigenmap(X2,F2);
%
%       % reorder and flip signs of E2
%       [E1,E2,Q] = reordereigenmaps(X1,X2,E1,E2,V1,V2);
%
%    See also eigenmap match aligneigenmaps
%
%    Reference:
%    [1] Spectral correspondence for Brain Matching,
%        Lombaert, Grady, Polimeni, Cheriet, IPMI'11

% file:    reordereigenmaps.m
% author:  Herve Lombaert
% contact: herve.lombaert@gmail.com
% date:    April 16th, 2011

function [E1,E2,Q] = reordereigenmaps(X1,X2,E1,E2,V1,V2)

k = size(E1,2); % number of eigenmaps

%% downsample data
idx1 = randperm(size(X1,1)); % randomly choose a few vertices
idx2 = randperm(size(X2,1));
idx1 = idx1(1:8:end); % downsample
idx2 = idx2(1:8:end);
LX1 = X1(idx1,:); % low res vertices
LE1 = E1(idx1,:); % low res eigenvector
LX2 = X2(idx2,:);
LE2 = E2(idx2,:);

%% Get pairs of closest points between two meshes

% scale X1 and X2 within a common box
LX1 = bsxfun(@rdivide, bsxfun(@minus,LX1,min(LX1)), range(LX1));
LX2 = bsxfun(@rdivide, bsxfun(@minus,LX2,min(LX2)), range(LX2));

% get closest points
tree = kdtree(LX2);
idx  = kdtree_closestpoint(tree, LX1);

%% Normalize eigenvector
for i = 1:k
    LE1(:,i) = (LE1(:,i)-min(LE1(:,i))) / (max(LE1(:,i))-min(LE1(:,i))) - 0.5;
    LE2(:,i) = (LE2(:,i)-min(LE2(:,i))) / (max(LE2(:,i))-min(LE2(:,i))) - 0.5;
end

%% Compute for each eigenmaps comparison cost and flip sign
C_spatial   = zeros(k-1); % cost of comparing E1(:,i) and E2(:,j)
C_spatial_f = zeros(k-1); % cost of comparing E1(:,i) and -E2(:,j)

for i = 1:k
    for j = 1:k
        % Eigenfunction diff between at this pair of closest points
        C_spatial  (i,j) = sum((LE2(idx,j) - LE1(:,i)).^2);

        % Eigenfunction diff between at this pair of closest points (with flipped E2)
        C_spatial_f(i,j) = sum((-LE2(idx,j) - LE1(:,i)).^2);
    end
end

C_spatial   = sqrt(C_spatial  ) / size(LX1,1);
C_spatial_f = sqrt(C_spatial_f) / size(LX1,1);

% Remember where are the sign flips ( S:1 if sign flip between E1(:,i) and E2(:,j) )
Q = min(C_spatial, C_spatial_f); % optimize with best cost
S = C_spatial > C_spatial_f;     % same sign: 0, flip sign: 1

[rows_flipped,cols_flipped] = find(S); % where are the flips

%% Find best matching
M = hungarian(Q);
[rows,cols] = find(M);

% Get the permutation costs
Q = sum(M'*Q')'; % cost for ith pair of eigenvectors

%% Flip signs
flipped = intersect([rows,cols],[rows_flipped,cols_flipped],'rows'); % which eigenmaps to flip
if ~isempty(flipped)
    E2(:,flipped(:,2)) = -E2(:,flipped(:,2));
end

%% Reorder E2
E2(:,rows) = E2(:,cols);
V2(rows)   = V2(cols);

end