%% displaycorrespondence.m
%% Herve Lombaert, April 28th, 2010
%%
%% Display two hemispheres and show its correspondences
%% Interactive marker: Click anywhere on the top hemisphere
%% and its corresponding point will be shown on the bottom hemisphere
%%
%% X1,F1 - first mesh
%% X2,F2 - second mesh
%% Correspondence - correspondence map ( corr(i) gives the corresponding point index )
%% E1,E2 - Eigenmaps of first and second meshes
%% C1,C2 - Colors to be displayed (default is fiedler vector)
%% nblinks - number of links to show (default: 30)
%% cam_phi,cam_theta - angle of camera
%%

function displaycorrespondence(X1,X2,F1,F2,Correspondence,E1,E2,C1,C2,nblinks,cam_phi,cam_theta)

if nargin<8;  C1        = E1(:,1); end
if nargin<9;  C2        = E2(:,1); end
if nargin<10; nblinks   = 30;      end
if nargin<11; cam_phi   = -50;     end;
if nargin<12; cam_theta = -30;     end;

%% Compute faces if not provided
if isempty(F1) || isempty(F2);
    F1 = MyCrustOpen([X1(:,1) X1(:,2) X1(:,3)]);
    F2 = MyCrustOpen([X2(:,1) X2(:,2) X2(:,3)]);
end

%% Callback whenever the mouse is clicked in data mode
h = datacursormode(gcf);
set(h,'UpdateFcn',@highlight,'SnapToDataVertex','on'); % callback function
datacursormode on;

data.marker = [];
data.X1   = X1;
data.F1   = F1;
data.X2   = X2;
data.F2   = F2;
data.Correspondence = Correspondence;
data.E1   = E1;
data.E2   = E2;
set(gcf,'UserData',data); % previous marker is empty

k = size(E1,2);
offsetx = 0;%max(range(X1(:,3))/2, range(X2(:,3))/2) * 1.5;
offsetz = max(range(X1(:,3))/2, range(X2(:,3))/2) * 1.1;

%% Display meshes using Fiedler vector as color
trisurf(F1,X1(:,1)+offsetx,X1(:,2),X1(:,3)+offsetz,C1);
daspect([1 1 1]); axis off; shading interp; %material dull;
%lighting gouraud;
hold on;
trisurf(F2,X2(:,1)-offsetx,X2(:,2),X2(:,3)-offsetz,C2);
daspect([1 1 1]); axis off; shading interp; %material dull;
%lighting gouraud;
hold off;
%camorbit(cam_phi,cam_theta);
repositioncam(X1,-0.5);
alpha(0.9);

txt = 'Click on the top cortex to find its match';
%text(0,0,2.2*offsetz,txt)

%% Display a few random links
displaylinks(nblinks);

    function displaylinks(nblinks)
        idx1 = randperm(size(X1,1)); % random links
        
        %% Display a few corresponding points
        for i=1:nblinks
            %% Find correspondence
            pos = X1(idx1(i),:);
            if isempty(Correspondence)
                [closest idx2] = min(sum((repmat(E1(idx1(i),2:k),size(X2,1),1)-E2(:,2:k))'.^2));
            else
                idx2 = Correspondence(idx1(i));
            end
            corr = X2(idx2,:);
            
            hold on;
            plot3([pos(1)+offsetx corr(1)-offsetx],[pos(2) corr(2)],[pos(3)+offsetz corr(3)-offsetz],'r-', pos(1)+offsetx,pos(2),pos(3)+offsetz,'ro', corr(1)-offsetx,corr(2),corr(3)-offsetz,'rv', 'MarkerSize',6); % draw a marker
            hold off;
        end
    end

    %% Get correspondence
    function corr = getCorrespondence(pos)
        data = get(gcf,'UserData');
        E1 = data.E1;
        E2 = data.E2;
        X1 = data.X1;
        X2 = data.X2;
        Correspondence = data.Correspondence;
        d = size(E1,2);
        k = size(E1,2);

        % where is the clicked point on mesh 1
        [closest idx1] = min(sum((repmat([pos(1)-offsetx pos(2) pos(3)-offsetz],size(X1,1),1)-X1)'.^2));
        
        % corresponding point on mesh 2
        if isempty(Correspondence)
            [closest idx2] = min(sum((repmat(E1(idx1,:),size(X2,1),1)-E2(:,:))'.^2));
        else
            idx2 = Correspondence(idx1);
        end
        corr = X2(idx2,:);
    end

    %% Callback function
    function [txt] = highlight(obj,event)
        pos = get(event,'Position'); % get position of cursor

        %% Delete previous marker
        data = get(gcf,'UserData'); % get previous marker
        if ~isempty(data.marker); delete(data.marker); end; % delete previous marker

        %% Get corresponding point
        corr = getCorrespondence(pos);
 
        %% Display marker text
        %txt = {[num2str(pos(1)) ', ' num2str(pos(2)) ', ' num2str(pos(3))]};
        txt = {['']};

        %% Display marker
        hold on;
        data.marker = plot3([pos(1) corr(1)-offsetx],[pos(2) corr(2)],[pos(3) corr(3)-offsetz],'g-', corr(1)-offsetx,corr(2),corr(3)-offsetz,'go', 'MarkerSize',6,'LineWidth',3); % draw a marker
        set(gcf,'UserData',data); % remember our marker
        hold off;
    end

end
